
import os
import re

import jwt
import uuid

import groups

GUEST = 0
USER = 1
STAFF = 128

JWT_SECRET = os.environ.get('ALLOC_JWT_SECRET', 'DEBUG_JWT_SECRET')
STAFF_PASSWORD = os.environ.get('ALLOC_STAFF_PW', 'PASSWORD')

_usercache = {}

class User(object):

    def __init__(self, uid = None):
        if uid is None:
            self.uid = uuid.uuid4()
        else:
            self.uid = uid
        self.name = ""
        self.role = GUEST
        self.sessions = set()

    @property
    def id(self):
        return str( self.uid )

    def has_access(self, level):
        return self.role >= level

    def set_name(self, name):
        self.name = re.sub(r'[^\w -]', '', name)

class UserManager(object):

    def __init__(self):
        self.cache = {}

    def exists(self, uid):
        return uid in self.cache

    def setup(self, data):
        tmp = User( data['uid'] )
        tmp.name = data.get('name', 'Unknown User')
        tmp.role = int( data.get('role', GUEST) )
        self.cache[ str(tmp.uid) ] = tmp
        return tmp

    def get(self, uid):
        return self.cache.get( str(uid), None )

    def get_name(self, uid):
        user = self.cache.get( str(uid), None )
        if not user:
            return "Unknown User"
        return user.name

    def get_sessions(self, uid):
        user = self.cache.get( str(uid), None )
        if not user:
            return set()
        return user.sessions

def handle_setup(server, user, message):
    role = USER

    pw = message.get('password', None)
    if pw:
        if pw == STAFF_PASSWORD:
            role = STAFF
        else:
            return None, None

    data = {
        'uid': str(uuid.uuid4()),
        'name': re.sub(r'[^\w -]', '', message.get('name', '')),
        'role': role
    }

    user = server.users.setup(data)
    ticket = jwt.encode( data, JWT_SECRET, algorithm="HS256" )
    return user, ticket


def handle_login(server, client, message):
    try:
        ticket = jwt.decode( message, JWT_SECRET, algorithms=["HS256"] )
        if 'uid' in ticket:
            user = server.users.get( ticket['uid'] )
            if not user:
                print("ticket had a UID, but that uid doesn't exist? - server restarted?")
                user = server.users.setup( ticket )
                return user, None
            return user, None
        else:
            return None, None
    except jwt.exceptions.InvalidSignatureError:
        return None, None

def handle_rename(server, client, message):
    if not client.user or not client.user.has_access(STAFF):
        print("non-admin tried force rename")
        return

    if 'uuid' not in message['data'] or 'name' not in message['data']:
        print("malformed force-rename message", message)
        return

    # get the user the staff member wishes to rename
    try:
        user_id = uuid.UUID( "{{{}}}".format(message['data']['uuid']) )
        user = server.users.get( user_id )
        if user is None:
            print("tried to rename unknown user")
            return

        user.set_name( message['data'].get('name', 'RENAMED USER') )
        # tell the user
        server.sendTopicMany( user.sessions, 'userRename', { 'uid': user.id, 'name': user.name })

        # tell the admins
        server.sendTopicMany( server.onlineStaff, 'userRenameAdm', { 'uid': user.id, 'name': user.name })

        # reissue group data (if any)
        group = server.groups.get_group( user.uid )
        if group is not None:
            sessions = []
            members = server.groups.members( group )
            for user in server.groups.members( group ):
                sessions += server.users.get(user).sessions
            server.sendTopicMany( sessions, 'group', groups.oneGroupJson( server, group, members ) )

    except ValueError:
        print("error parsing UUID")
