
import random
import itertools
import logging
from collections import defaultdict

import auth

def lMin( items ):
    idx = 0
    val = items[0]
    for i in range(1, len(items)):
        if val < items[i]:
            idx = i
            val = items[i]
    return idx

class GroupManager(object):

    def __init__(self):
        self.groups = []
        self.maxGroups = None

    def _allocateVariable(self, students, idealSize=2, minMembers=1):
        """Create groups based on an ideal size.

        The number of groups is variable."""
        if not students:
            return []

        random.shuffle( students )

        # group allocations
        groups = [ students[n:n+idealSize] for n in range(0, len(students), idealSize) ]

        # less than idealSize students - 1 group
        if not groups:
            return [ students ]

        nGroups = len(groups)

        # deal with the last group
        if len( groups[-1] ) < minMembers and len(groups) < 1:
            lastGroup = groups.pop()
            nGroups -= 1
            self._allocateUniform( lastGroup, nGroups )
        return groups

    def _allocateFixed(self, students, idealGroups=4, minMembers=2):
        """Create a fixed number of groups and allocate students accordingly"""
        maxPossibleGroups = len( students ) // minMembers
        numGroups = min( idealGroups, maxPossibleGroups )

        # there must be at least one group
        if numGroups == 0:
            numGroups = 1

        studentsPerGroup = len(students) // numGroups
        return self._allocateVariable( students, studentsPerGroup, minMembers)

    def _allocateUniform(self, people, n=None):
        people = list( people )
        if n is None:
            n = len( self.groups )

        idx = 0
        while people:
            groups[ groupAlloc ].append( people.pop() )
            groupAlloc = (idx + 1) % n
        return people

    def _allocateSmallest(self, people):
        while people:
            smallestGroup = lMin( self.groups )
            self.groups[smallestGroup].append( people.pop() )

    @property
    def everyone(self):
        return set( itertools.chain( self.groups ) )

    def members(self, group):
        return self.groups[ group ]

    def find(self, member):
        for i in range(0, len(self.groups)):
            if member in self.groups[i]:
                return i
        return None

    def allocate(self, students, method, idealNumber, minMembers=2):
        """Do a full reallocation, ignoring any current team allocations"""
        if method == 'members':
            self.method = method
            self.groups = self._allocateVariable(students, idealNumber, minMembers)
            return self.groups
        elif method == 'groups':
            self.method = method
            self.groups = self._allocateFixed(students, idealNumber, minMembers)
            return self.groups
        else:
            logging.warning("got unknown group allocation method: {}", method)
            return []

    def get_group(self, uid):
        for idx in range(0, len(self.groups)):
            if uid in self.groups[idx]:
                return idx
        return None

    def remove(self, people, minMembers):
        # remove missing people from allocation
        extra_spaces = defaultdict(int)
        illegal_groups = set()
        empty_groups = set()

        for person in people:
            groupID = self.find( person )
            self.groups[ groupID ].remove( person )
            extra_spaces[ groupID] += 1

            if len( self.groups[groupID] ) == 0:
                empty_groups.append( groupID )
            elif len( self.groups[groupID] ) < minMembers:
                illegal_groups.append( groupID )

        return extra_spaces, illegal_groups, empty_groups

    def update(self, students, method):
        """Do a partial reallocation, taking into account existing teams"""
        new_people = list(set( students ) - self.everyone)
        missing_people = list(self.everyone - set(students))

        if missing_people:
            spaces, illegal, empty = self.remove( missing_people, minMembers )

        if method == 'uniform':
            self._allocateUniform( new_people )
        elif method == 'smallest':
            self._allocateSmallest( new_people )

def oneGroupJson(server, teamID, members):
    return {
            'id': teamID,
            'members': [ str( member ) for member in members ],
            'names': [ server.users.get_name( member ) for member in members ]
        }

def groupJson(server, teams):
    all_teams = []
    for teamID in range(0, len(teams)):
        members = teams[ teamID ]
        payload = oneGroupJson( server, teamID, members )
        all_teams.append( payload )
    return all_teams

def handle_allocate( server, client, message ):
    if client.role != auth.STAFF:
        logging.warning("{} tried to use allocate, but is not staff", client.id)
        return

    # parameters
    data = message.get('data', {})
    idealSize = int(data.get('idealSize', 4))
    minSize = int(data.get('minSize', 2))
    method = data.get('method', 'members')

    # allocation
    active_sessions = server.active( lambda x: x.role == auth.USER )
    users = list(set( [ server.clients[session].user.uid for session in active_sessions ] ))
    teams = server.groups.allocate( users, method, idealSize, minSize )

    all_teams = groupJson( server, teams )

    # send team allocations
    for teamID in range(0, len(teams)):
        payload = all_teams[ teamID ]

        sessions = set()
        for member in teams[ teamID ]:
            sessions = sessions.union( server.users.get_sessions( member ) )
        server.sendTopicMany( sessions, 'group', payload )

    server.sendTopicMany( server.onlineStaff, 'groups', all_teams )
