
import websockets
import asyncio
import uuid
import json

import auth
import groups
import util

async def send(websocket, message):
    try:
        await websocket.send( message )
    except websocket.ConnectionClosed:
        pass

class ClientHandler(object):

    def __init__(self, ws):
        self.id_ = uuid.uuid4()
        self.ws = ws
        self.user = None
        self.queue = asyncio.Queue()

    @property
    def name(self):
        if not self.user:
            return "Unknown User"
        else:
            return self.user.name

    @property
    def role(self):
        if not self.user:
            return auth.GUEST
        else:
            return self.user.role

    async def sendNow(self, topic, message):
        msg = {
            'type': topic,
            'data': message
        }
        await self.ws.send( json.dumps( msg ) )

    def send(self, message):
        self.queue.put_nowait( message )

    async def login(self, server):
        try:
            await self.sendNow( 'welcome', {
                'id': self.id
            })

            msg = json.loads(await self.ws.recv())
            if 'type' not in msg:
                return None

            print( msg )
            if msg['type'] == 'ticket':
                print("ticket authentication")
                user, ticket = auth.handle_login(server, self, msg['data'])
                if not user:
                    await self.ws.close(1011, "invalid ticket")
                    return None
            elif msg['type'] == 'setup':
                print("user setup")
                user, ticket = auth.handle_setup(server, self, msg['data'])
                if not ticket:
                    await self.ws.close(1011, "authentication failed")
                    return None
            else:
                print("invalid handshake")
                await self.ws.close(1011, "invalid handshake")
                return None

            self.user = user
            group = server.groups.find( user.uid )

            await self.sendNow( 'user', {
                'uid': user.uid,
                'name': user.name,
                'role': user.role,
                'group': server.groups.find( user.uid ),
                'ticket': ticket
            })

            if group is not None:
                await self.sendNow('group', groups.oneGroupJson( server, group, server.groups.members( group ) ) )

            if user.has_access( auth.STAFF ):
                onlineUsers = [ util.statusPayload(c) for c in server.clients.values() ]
                await self.sendNow( 'users', onlineUsers )
                onlineGroups = groups.groupJson( server, server.groups.groups )
                await self.sendNow( 'groups', onlineGroups )

            return user
        except websockets.exceptions.ConnectionClosed:
            return None

    async def consumer(self):
        while True:
            message = await self.queue.get()
            print("[>>]", self.id, message)
            await send( self.ws, json.dumps(message) )

    async def producer(self, server):
        async for messageTxt in self.ws:
            message = json.loads( messageTxt )
            print("[<<]", self.id, message)
            await server.process(self, message)

    async def handler(self, server):
        consumer_task = asyncio.create_task( self.consumer() )
        producer_task = asyncio.create_task( self.producer( server ) )
        done, pending = await asyncio.wait(
            [consumer_task, producer_task],
            return_when=asyncio.FIRST_COMPLETED,
        )
        for task in pending:
            task.cancel()

    @property
    def id(self):
        return str(self.id_)

class SocketServer(object):

    def __init__(self, callbacks):
        self.clients = {}
        self.protocol = callbacks

    async def handler(self, ws):
        client = ClientHandler(ws)

        # Handshake
        user = await client.login(self)
        if not user:
            return

        # main loop
        self.clients[ client.id_ ] = client
        client.user.sessions.add( client.id_ )
        self.sendTopicMany( self.onlineStaff, 'status', util.statusPayload( client ))
        await client.handler( self )
        client.user.sessions.remove( client.id_ )
        del self.clients[ client.id_ ]
        self.sendTopicMany( self.onlineStaff, 'status', util.statusPayload(client, False) )

    async def process(self, client, message):
        if 'type' not in message:
            return

        topic = message.get('type', None)
        if topic in self.protocol:
            self.protocol[ topic ]( self, client, message )

    @property
    def onlineStaff(self):
        return [ x.id_ for x in self.clients.values() if x.user.role == auth.STAFF ]

    def activeID(self, f=None):
        """Return active session UUIDs (as strings)"""
        if f is None:
            return [ str(k) for k in self.clients.keys() ]
        else:
            return [ v.id for (k,v) in self.clients.items() if f(v) ]

    def active(self, f=None):
        """Return active session UUIDs (as UUIDs)"""
        if f is None:
            return [ k for k in self.clients.keys() ]
        else:
            return [ v.id_ for (k,v) in self.clients.items() if f(v) ]

    def sendTopic(self, client, topic, payload):
        self.send(client, {
            'type': topic,
            'data': payload
        })

    def sendTopicMany(self, clients, topic, payload):
        self.sendToMany(clients, {
            'type': topic,
            'data': payload
        })

    def sendTopicAll(self, topic, payload):
        self.broadcast({
            'type': topic,
            'data': payload
        })

    def send(self, client, message):
        self.clients[ client ].send( message )

    def sendToMany(self, clients, message):
        for cid in clients:
            if cid in self.clients:
                websocket = self.clients[ cid ]
                websocket.send( message )

    def broadcast(self, message):
        for websocket in self.clients.values():
            websocket.send( message )

def handle_notify( server, client, message ):
    if not client.user or not client.user.has_access( auth.STAFF ):
        return

    target = server.users.get( message['data']['uuid'] )
    if not target:
        print("no idea who that is...")
        return

    if target.sessions:
        server.sendTopicMany( target.sessions, 'notify', { 'message': message['data']['message'] } )

def handle_kick( server, client, message ):
    """Note this doesn't actually kick the client, it just asks it to disconnect.
    I could actually force disconnect, blacklist the session and user IDs, but that's a little overkill.

    It's mostly useful for force-invalidating sessions."""
    if not client.user or not client.user.has_access( auth.STAFF ):
        return

    target = server.users.get( message['data']['uuid'] )
    if not target:
        print("no idea who that is...")
        return

    if target.sessions:
        server.sendTopicMany( target.sessions, 'kick', {} )

