#! /usr/bin/env python3

from collections import namedtuple
from websockets import serve

import jwt

import json
import asyncio
import random
import uuid
import re
import os

import alloc

#
# todo list
#
# * cache groups in persistant storage
# * allow students not in groups to be added to existing allocation
# * better management of joining/leaving students
# * add support for topics
# * add task timer for groups

# FIXME: should support _FILE for secrets
JWT_SECRET = os.environ.get('ALLOC_JWT_SECRET', 'DEBUG_JWT_SECRET')
STAFF_PASSWORD = os.environ.get('ALLOC_STAFF_PW', 'PASSWORD')

GUEST = 0
STUDENT = 1
STAFF = 2

LEGAL_TOPICS = set(['connect', ])

# caches
LAST_GROUPS = []
LAST_NAMES = {}

def allocateVariableGroups(students, idealTeamSize=4, minTeamSize=2):
    """Allocate students to groups, with an ideal team size.
    If there are students left over, and there are at least minStudents they get
    to be a team. Else, the 'remainder' students are distrubted across teams.
    """
    random.shuffle( students )
    groups = []

    # Allocate teams to groups
    gid = 0
    currGroup = []
    for student in students:
        currGroup.append( student )
        if idealTeamSize == len(currGroup):
            groups.append( currGroup )
            currGroup = []
            gid += 1

    # last team
    if currGroup:
        if minTeamSize <= len(currGroup) or len(groups) == 0:
            # viable team (or no other choice)
            groups.append( currGroup )
        else:
            # redistribute across teams
            groupAlloc = 0
            while currGroup:
                groups[groupAlloc].append( currGroup.pop() )
                groupAlloc = (groupAlloc + 1) % len(groups)

    return groups

def allocateFixedGroups(students, numGroups, minSize = 1):
    """Allocate students assuming a fixed number of groups.

    Students are allocated fairly amoung groups, remainders
    will be added to the first n groups. If there are not
    enouph students for numGroups, then the maximum number
    of groups to still satify numGroups is used.
    """

    # figure out how many groups we can have
    maxPossibleGroups = len(students) // minSize
    numGroups = min( numGroups, maxPossibleGroups )

    if numGroups == 0:
        numGroups = 1 # we're having a bad day

    groups = []
    for i in range(0, numGroups):
        groups.append( [] )

    random.shuffle( students )
    nextGroup = 0
    while students:
        groups[nextGroup].append( students.pop() )
        nextGroup = (nextGroup + 1) % numGroups

    return groups

ALLOCATIONS = {
    'members': allocateVariableGroups,
    'group': allocateFixedGroups
}

async def sendToStaff( topic, data ):
    for staff in get_staff():
        await clients[staff].send_topic( topic, data )

async def changeName(client, message):
    realName = message['data']['name']
    realName = re.sub(r'[^\w -]', '', realName)

    if 'password' in message['data']:
        if message['data']['password'] == STAFF_PASSWORD:
            client.role = STAFF
        else:
            client.role = GUEST
    else:
        client.role = STUDENT

    await client.rename( realName )

async def forceAllocate(client, message):
    if client.role != STAFF:
        return

    data = message["data"]

    # Allow different group sizes
    size = 4
    if 'idealSize' in data:
        size = int(data['idealSize'])

    minSize = 2
    if 'minSize' in data:
        minSize = int(data['minSize'])
    
    method = allocateFixedGroups
    if 'method' in data and data['method'] in ALLOCATIONS:
        method = ALLOCATIONS[ data['method'] ]

    summary = {
        'size': size,
        'groups': []
    }

    # Perform group allocation
    ids = [ x for x in clients.keys() if clients[x].role == STUDENT ]
    groups = method( ids, size, minSize )

    # send out group notifications
    for (idx, group) in enumerate(groups):
        members = [ str( clients[x].name ) for x in group ]
        data = {
            'groupID': idx,
            'members': members
        }
        summary['groups'].append( data )
        for student in group:
            await clients[ student ].send_topic( 'allocate', data )

    # Send master list to staff
    LAST_GROUPS = summary
    await sendToStaff( 'groups', summary )

async def resumeToken(client, message):
    try:
        ticket = jwt.decode( message['data'], JWT_SECRET, algorithms=["HS256"] )
        resumed_token = uuid.UUID("{"+ticket['uuid']+"}")
        del clients[ client.uuid ]
        print( "client {} used jwt to reclaim id {}", client.uuid, resumed_token )

        # resume session
        client.uuid = resumed_token
        clients[ client.uuid ] = client
        if resumed_token in LAST_NAMES:
            session = LAST_NAMES[ resumed_token ]
            client.name = session[0]
            client.role = session[1]
            await client.send_topic('changeName', { 'name': client.name, 'role': client.role })

            if client.role == STAFF:
                students = [ {'uuid': str(x.uuid), 'name': x.name} for x in clients.values() if x.role == STUDENT ]
                await client.send_topic('students', students)

                if LAST_GROUPS:
                    await client.send_topic('groups', LAST_GROUPS)
    except jwt.exceptions.DecodeError:
        pass

async def forceRename(client, message):
    if client.role != STAFF:
        return

    realName = message['data']['name']
    realName = re.sub(r'[^\w -]', '', realName)
    client_token = uuid.UUID("{"+message['data']['uuid']+"}")

    await clients[ client_token ].rename( realName )

protocol = {
        'changeName': changeName,
        'token': resumeToken,

        # staff options
        'forceAllocate': forceAllocate,
        'forceRename': forceRename,
}

class Client:

    def __init__(self, session, socket, role = GUEST):
        self.uuid = session
        self.name = None
        self.role = role
        self.socket = socket

    async def rename(self, new_name):
        self.name = new_name
        LAST_NAMES[ self.uuid ] = ( new_name, self.role )

        # staff annoucment
        sendToStaff('nameChanged', { 'uuid': str(self.uuid), 'name': new_name } )
        await self.send_topic('changeName', {'name': new_name, 'role': self.role })

    async def process(self):
        async for message in self.socket:
            data = json.loads( message )
            msg_type = data['type']
            if msg_type in protocol:
                await self._recv_topic( data )

    async def send_topic(self, topic, data):
        msg = {
            'type': topic,
            'data': data
        }
        print( "[>>]", msg )
        await self.socket.send( json.dumps(msg) )

    async def recv_topic(self, data):
        print("[<<]", data)
        msg_type = data['type']
        if msg_type in protocol:
            await protocol[ msg_type ]( self, data )

class Room(object):

    def __init__(self, server):
        self._server = server
        self.name = ""
        self._sessions = {}
        self._groups = {}

    @property
    def staff(self):
        return [ k for (k,v) in self._sessions if v.role == STAFF ]

    @property
    def students(self):
        return [ k for (k,v) in self._sessions if v.role == STUDENT ]

    @property
    def groups(self):
        return self._groups.keys()

    def group(self, group):
        if group not in self._groups:
            return []
        else:
            return self._groups[group]

    def join(self, session):
        self._sessions[session.uuid] = session

    def part(self, session):
        del self._sessions[session.uuid]

    async def sendStaff(self, topic, payload):
        await self._sendBulk( self.staff, topic, payload )

    async def sendStudents(self, topic, payload):
        await self._sendBulk( self.students, topic, payload )

    async def sendGroup(self, group, topic, payload):
        await self._sendBulk( self._groups[group], topic, payload )

    async def _sendBulk(self, people, topic, payload):
        jobs = [ self._server.send( uuid, topic, payload)
                 for uuid
                 in people
                ]
        return asyncio.gather( jobs )

    async def sendTo(self, uuid, topic, payload):
        if uuid in self._sessions:
            await self._sessions[ uuid ].send_topic( token, payload )


class Server(object):

    def __init__(self):
        self._rooms = {}
        self._clients = {}

    def create(self, room):
        self._rooms[room] = Room(self)

    async def send(self, client, topic, payload = None):
        if payload is None:
            payload = {}
        if topic not in LEGAL_TOPICS:
            print("unknown topic type", topic)
        if client in self._clients:
            await self._clients[ client ].send_topic( topic, payload )
            return True
        return False

    async def handleClient(self, websocket):
        token = await websocket.recv()

        # process token
        ticket = jwt.decode( token, JWT_SECRET, algorithms=["HS256"] )
        resumed_token = uuid.UUID("{"+ticket['uuid']+"}")

        # force disconnect
        #if resumed_token in self._clients:
        #    self._clients[resumed_token].disconnect()

        # create client
        client = Client(resumed_token, websocket)
        self._clients[ resumed_token ] = client

        await self.send( client.uuid, 'connect')
        await client.process()
        del self._clients[ client.uuid ]

server = Server()
server.create( "main" )

clients = {}

def get_students():
    return [ x for x in clients.keys() if clients[x].role == STUDENT ]

def get_staff():
    return [ x for x in clients.keys() if clients[x].role == STAFF ]

async def handleClient(websocket):
    client = Client(websocket)

    clients[ client.uuid ] = client
    sendToStaff('clientJoin', { 'uuid': str(self.uuid) } )

    auth_token = jwt.encode( {'uuid': str(client.uuid)}, JWT_SECRET, algorithm="HS256" )
    await client.send_topic("token", auth_token)

    async for message in websocket:
        data = json.loads( message )
        msg_type = data['type']
        if msg_type in protocol:
            await client.recv_topic( data )

    del clients[ client.uuid ]
    sendToStaff('clientLeft', { 'uuid': str(self.uuid) } )

async def main():
    async with serve( server.handleClient, "0.0.0.0", 8081):
        await asyncio.Future()

asyncio.run( main() )
